@@ -1,7 +1,7 @@
 /*
  *  xfdesktop - xfce4's desktop manager
  *
- *  Copyright (c) 2004 Brian Tarricone, <bjt23@cornell.edu>
+ *  Copyright (c) 2004-2009 Brian Tarricone, <bjt23@cornell.edu>
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -32,6 +32,10 @@
 #include <stdlib.h>
 #endif
 
+#ifdef HAVE_UNISTD_H
+#include <unistd.h>
+#endif
+
 #ifdef HAVE_STRING_H
 #include <string.h>
 #endif
@@ -40,37 +44,38 @@
 #include <signal.h>
 #endif
 
+#ifdef HAVE_UNISTD_H
+#include <unistd.h>
+#endif
+
 #include <X11/Xlib.h>
 
-#include <gmodule.h>
 #include <gdk/gdkx.h>
 #include <gtk/gtk.h>
 
-#include <libxfce4mcs/mcs-client.h>
-#include <libxfcegui4/libxfcegui4.h>
+#include <xfconf/xfconf.h>
+#include <libxfce4ui/libxfce4ui.h>
 
 #ifdef ENABLE_FILE_ICONS
 #include <dbus/dbus-glib.h>
 #endif
 
-#include "main.h"
 #include "xfdesktop-common.h"
 #include "xfce-backdrop.h"
 #include "xfce-desktop.h"
-#include "xfce-desktop-settings.h"
 #include "menu.h"
 #include "windowlist.h"
-#include "settings.h"
 
-static SessionClient *client_session = NULL;
-static gboolean is_session_managed = FALSE;
-static gboolean desktop_gtk_menu_images = TRUE;
+#ifdef HAVE_LIBNOTIFY
+#include "xfdesktop-notify.h"
+#endif
+
+static XfceSMClient *sm_client = NULL;
 
 static void
-session_logout()
+session_logout(void)
 {
-    g_return_if_fail(is_session_managed);
-    logout_session(client_session);
+    xfce_sm_client_request_shutdown(sm_client, XFCE_SM_CLIENT_SHUTDOWN_HINT_ASK);
 }
 
 static void
@@ -79,18 +84,6 @@ session_die(gpointer user_data)
     gtk_main_quit();
 }
 
-static void
-desktop_settings_menu_images_notify(GObject *gobject,
-                                    GParamSpec *arg1,
-                                    gpointer user_data)
-{
-    g_object_get(G_OBJECT(gobject),
-                 "gtk-menu-images", &desktop_gtk_menu_images,
-                 NULL);
-    windowlist_set_show_icons(desktop_gtk_menu_images);
-    menu_set_show_icons(desktop_gtk_menu_images);
-}
-
 static void
 event_forward_to_rootwin(GdkScreen *gscreen, GdkEvent *event)
 {
@@ -164,50 +157,46 @@ scroll_cb(GtkWidget *w, GdkEventScroll *evt, gpointer user_data)
     return TRUE;
 }
 
-static gboolean
-button_cb(GtkWidget *w, GdkEventButton *evt, gpointer user_data)
-{
-    GdkScreen *gscreen = gtk_widget_get_screen(w);
-    gint button = evt->button;
-    gint state = evt->state;
-    
-    if(evt->type == GDK_BUTTON_PRESS) {
-        if(button == 2 || (button == 1 && (state & GDK_SHIFT_MASK)
-                           && (state & GDK_CONTROL_MASK)))
-        {
-            popup_windowlist(gscreen, button, evt->time);
-            return TRUE;
-        } else if(button == 3 || (button == 1 && (state & GDK_SHIFT_MASK))) {
-            popup_desktop_menu(gscreen, 0, evt->time);
-            return TRUE;
-        }
-    }
-    
-    return FALSE;
-}
-
 static gboolean 
 reload_idle_cb(gpointer data)
 {
-    settings_reload_all();
+    XfceDesktop **desktops = data;
+    gint i, nscreens;
+
+    nscreens = gdk_display_get_n_screens(gdk_display_get_default());
+    for(i = 0; i < nscreens; ++i) {
+        if(desktops[i])
+            xfce_desktop_refresh(desktops[i]);
+    }
+
     menu_reload();
+
     return FALSE;
 }
 
-gboolean
+static gboolean
 client_message_received(GtkWidget *w, GdkEventClient *evt, gpointer user_data)
 {
     if(evt->data_format == 8) {
         if(!strcmp(RELOAD_MESSAGE, evt->data.b)) {
-            g_idle_add ((GSourceFunc)reload_idle_cb, NULL);
+            g_idle_add((GSourceFunc)reload_idle_cb, user_data);
             return TRUE;
         } else if(!strcmp(MENU_MESSAGE, evt->data.b)) {
-            popup_desktop_menu(gtk_widget_get_screen(w), 0, GDK_CURRENT_TIME);
+            xfce_desktop_popup_root_menu(XFCE_DESKTOP(w), 0,
+                                         GDK_CURRENT_TIME);
             return TRUE;
         } else if(!strcmp(WINDOWLIST_MESSAGE, evt->data.b)) {
-            popup_windowlist(gtk_widget_get_screen(w), 0, GDK_CURRENT_TIME);
+            xfce_desktop_popup_secondary_root_menu(XFCE_DESKTOP(w), 0,
+                                                   GDK_CURRENT_TIME);
             return TRUE;
+#ifdef ENABLE_FILE_ICONS
+        } else if(!strcmp(ARRANGE_MESSAGE, evt->data.b)) {
+            xfce_desktop_arrange_icons(XFCE_DESKTOP(w));
+            return TRUE;
+#endif
         } else if(!strcmp(QUIT_MESSAGE, evt->data.b)) {
+            xfce_sm_client_set_restart_style(sm_client,
+                                             XFCE_SM_CLIENT_RESTART_NORMAL);
             gtk_main_quit();
             return TRUE;
         }
@@ -217,41 +206,75 @@ client_message_received(GtkWidget *w, GdkEventClient *evt, gpointer user_data)
 }
 
 static void
-sighandler_cb(int sig)
+xfdesktop_handle_SIGUSR1(gint sig,
+                         gpointer user_data)
 {
-    switch(sig) {
-        case SIGUSR1:
-            g_idle_add ((GSourceFunc)reload_idle_cb, NULL);
-            break;
-        
-        default:
-            gtk_main_quit();
-            break;
-    }
+    //settings_reload_all();  /* FIXME */
+    menu_reload();
+}
+
+static void
+xfdesktop_handle_quit_signals(gint sig,
+                              gpointer user_data)
+{
+    gint main_level;
+    
+    for(main_level = gtk_main_level(); main_level > 0; --main_level)
+        gtk_main_quit();
 }
 
 int
 main(int argc, char **argv)
 {
+    GOptionContext *octx;
     GdkDisplay *gdpy;
-    GtkWidget **desktops, *dummy;
+    GtkWidget **desktops;
     gint i, nscreens;
     Window xid;
-    McsClient *mcs_client;
-    GtkSettings *settings;
+    XfconfChannel *channel = NULL;
     const gchar *message = NULL;
     gboolean already_running;
+    gchar buf[1024];
+    GError *error = NULL;
+    gboolean opt_version = FALSE, opt_reload = FALSE;
+    gboolean opt_menu = FALSE, opt_windowlist = FALSE;
+    gboolean opt_arrange = FALSE, opt_quit = FALSE;
+    const GOptionEntry main_entries[] = {
+        { "version", 'V', G_OPTION_FLAG_IN_MAIN, G_OPTION_ARG_NONE, &opt_version, N_("Display version information"), NULL },
+        { "reload", 0, G_OPTION_FLAG_IN_MAIN, G_OPTION_ARG_NONE, &opt_reload, N_("Reload all settings, refresh image list"), NULL },
+        { "menu", 0, G_OPTION_FLAG_IN_MAIN, G_OPTION_ARG_NONE, &opt_menu, N_("Pop up the menu (at the current mouse position)"), NULL },
+        { "windowlist", 0, G_OPTION_FLAG_IN_MAIN, G_OPTION_ARG_NONE, &opt_windowlist, N_("Pop up the window list (at the current mouse position)"), NULL },
+#ifdef ENABLE_FILE_ICONS
+        { "arrange", 0, G_OPTION_FLAG_IN_MAIN, G_OPTION_ARG_NONE, &opt_arrange, N_("Automatically arrange all the icons on the desktop"), NULL },
+#endif
+        { "quit", 0, G_OPTION_FLAG_IN_MAIN, G_OPTION_ARG_NONE, &opt_quit, N_("Cause xfdesktop to quit"), NULL },
+        { NULL, 0, 0, 0, NULL, NULL, NULL }
+    };
     
     /* bind gettext textdomain */
     xfce_textdomain(GETTEXT_PACKAGE, LOCALEDIR, "UTF-8");
     
-#ifdef ENABLE_FILE_ICONS
+#if defined(ENABLE_FILE_ICONS) || defined(USE_DESKTOP_MENU)
     g_thread_init(NULL);
+#endif
+#ifdef ENABLE_FILE_ICONS
     dbus_g_thread_init();
 #endif
-    gtk_init(&argc, &argv);
-        
-    if(argc > 1 && (!strcmp(argv[1], "--version") || !strcmp(argv[1], "-V"))) {
+
+    octx = g_option_context_new("");
+    g_option_context_add_main_entries(octx, main_entries, NULL);
+    g_option_context_add_group(octx, gtk_get_option_group(TRUE));
+    g_option_context_add_group(octx, xfce_sm_client_get_option_group(argc, argv));
+
+    if(!g_option_context_parse(octx, &argc, &argv, &error)) {
+        g_printerr(_("Failed to parse arguments: %s\n"), error->message);
+        g_error_free(error);
+        return 1;
+    }
+
+    g_option_context_free(octx);
+
+    if(opt_version) {
         g_print(_("This is %s version %s, running on Xfce %s.\n"), PACKAGE,
                 VERSION, xfce_version_string());
         g_print(_("Built with GTK+ %d.%d.%d, linked with GTK+ %d.%d.%d."),
@@ -282,44 +305,28 @@ main(int argc, char **argv)
                 );
         
         return 0;
-    }
-
-    if(argc > 1) {
-        const gchar *argument = argv[1];
-        
-        /* allow both --(option) and -(option) */
-        if('-' == argument[0] && '-' == argument[1])
-            ++argument;
-        
-        if(!strcmp("-reload", argument))
-            message = RELOAD_MESSAGE;
-        else if(!strcmp("-menu", argument))
-            message = MENU_MESSAGE;
-        else if(!strcmp("-windowlist", argument))
-            message = WINDOWLIST_MESSAGE;
-        else if(!strcmp("-quit", argument))
-            message = QUIT_MESSAGE;
-        else if(!strcmp("--sm-client-id", argv[1])) {
-            /* do nothing */
-        } else {
-            g_printerr(_("%s: Unknown option: %s\n"), PACKAGE, argv[1]);
-            g_printerr(_("Options are:\n"));
-            g_printerr(_("    --reload      Reload all settings, refresh image list\n"));
-            g_printerr(_("    --menu        Pop up the menu (at the current mouse position)\n"));
-            g_printerr(_("    --windowlist  Pop up the window list (at the current mouse position)\n"));
-            g_printerr(_("    --quit        Cause xfdesktop to quit\n"));
-            
-            return 1;
-        }
-    }
+    } else if(opt_reload)
+        message = RELOAD_MESSAGE;
+    else if(opt_menu)
+        message = MENU_MESSAGE;
+    else if(opt_windowlist)
+        message = WINDOWLIST_MESSAGE;
+    else if(opt_arrange)
+        message = ARRANGE_MESSAGE;
+    else if(opt_quit)
+        message = QUIT_MESSAGE;
     
+    signal(SIGPIPE, SIG_IGN);
+
     already_running = xfdesktop_check_is_running(&xid);
     if(already_running) {
-        DBG("xfdesktop is running");
-        if(!message)
+        if(!message) {
+            g_printerr("%s[%d] is already running; assuming --reload\n",
+                       PACKAGE, (int)getpid());
             message = RELOAD_MESSAGE;
+        }
     }
-    
+
     if(message) {
         if(!already_running)
             g_printerr(_("%s is not running.\n"), PACKAGE);
@@ -328,69 +335,74 @@ main(int argc, char **argv)
 
         return (already_running ? 0 : 1);
     }
-    
+
+    DBG ("starting up");
+
+    sm_client = xfce_sm_client_get();
+    xfce_sm_client_set_restart_style(sm_client, XFCE_SM_CLIENT_RESTART_IMMEDIATELY);
+    xfce_sm_client_set_priority(sm_client, XFCE_SM_CLIENT_PRIORITY_DESKTOP);
+    g_signal_connect(sm_client, "quit",
+                     G_CALLBACK(session_die), NULL);
+
+    if(!xfce_sm_client_connect(sm_client, &error) && error) {
+        g_printerr("Failed to connect to session manager: %s\n", error->message);
+        g_clear_error(&error);
+    }
+
     gdpy = gdk_display_get_default();
-    mcs_client = settings_init();
-    
-    /* need GtkImageMenuItem to install the property */
-    dummy = gtk_image_menu_item_new();
-    gtk_widget_destroy(dummy);
-    
-    settings = gtk_settings_get_for_screen(gdk_display_get_default_screen(gdpy));
-    g_object_get(G_OBJECT(settings),
-                 "gtk-menu-images", &desktop_gtk_menu_images,
-                 NULL);
-    g_signal_connect(G_OBJECT(settings), "notify::gtk-menu-images",
-                     G_CALLBACK(desktop_settings_menu_images_notify), NULL);
-    
+
+    if(!xfconf_init(&error)) {
+        g_warning("%s: unable to connect to settings daemon: %s.  Defaults will be used",
+                  PACKAGE, error->message);
+        g_error_free(error);
+        error = NULL;
+    } else
+        channel = xfconf_channel_get(XFDESKTOP_CHANNEL);
+
     nscreens = gdk_display_get_n_screens(gdpy);
-    desktops = g_new(GtkWidget *, nscreens);
+    desktops = g_new0(GtkWidget *, nscreens);
     for(i = 0; i < nscreens; i++) {
-        desktops[i] = xfce_desktop_new(gdk_display_get_screen(gdpy, i));
+        g_snprintf(buf, sizeof(buf), "/backdrop/screen%d/", i);
+        desktops[i] = xfce_desktop_new(gdk_display_get_screen(gdpy, i),
+                                       channel, buf);
         gtk_widget_add_events(desktops[i], GDK_BUTTON_PRESS_MASK
                               | GDK_BUTTON_RELEASE_MASK | GDK_SCROLL_MASK);
         g_signal_connect(G_OBJECT(desktops[i]), "scroll-event",
                          G_CALLBACK(scroll_cb), NULL);
-        g_signal_connect_after(G_OBJECT(desktops[i]), "button-press-event",
-                               G_CALLBACK(button_cb), NULL);
-        if(mcs_client) {
-            settings_register_callback(xfce_desktop_settings_changed, desktops[i]);
-            xfce_desktop_settings_load_initial(XFCE_DESKTOP(desktops[i]),
-                                               mcs_client);
-        }
+        g_signal_connect(G_OBJECT(desktops[i]), "client-event",
+                         G_CALLBACK(client_message_received), desktops);
+        menu_attach(XFCE_DESKTOP(desktops[i]));
+        windowlist_attach(XFCE_DESKTOP(desktops[i]));
         gtk_widget_show(desktops[i]);
         gdk_window_lower(desktops[i]->window);
     }
     
-    signal(SIGPIPE, SIG_IGN);
-    
-    client_session = client_session_new(argc, argv, NULL,
-                                        SESSION_RESTART_IF_RUNNING, 35);
-    client_session->die = session_die;
-    is_session_managed = session_init(client_session);
-    
-    if(is_session_managed) {
-        for(i = 0; i < nscreens; ++i)
-            xfce_desktop_set_session_logout_func(XFCE_DESKTOP(desktops[i]),
-                                                 session_logout);
+    for(i = 0; i < nscreens; ++i) {
+        xfce_desktop_set_session_logout_func(XFCE_DESKTOP(desktops[i]),
+                                             session_logout);
     }
+
+    menu_init(channel);
+    windowlist_init(channel);
     
-    menu_init(mcs_client);
-    menu_set_show_icons(desktop_gtk_menu_images);
-    
-    windowlist_init(mcs_client);
-    windowlist_set_show_icons(desktop_gtk_menu_images);
-    
-    if(mcs_client) {
-        settings_register_callback(menu_settings_changed, desktops[0]);
-        settings_register_callback(windowlist_settings_changed, desktops[0]);
+    if(xfce_posix_signal_handler_init(&error)) {
+        xfce_posix_signal_handler_set_handler(SIGHUP,
+                                              xfdesktop_handle_quit_signals,
+                                              desktops, NULL);
+        xfce_posix_signal_handler_set_handler(SIGINT,
+                                              xfdesktop_handle_quit_signals,
+                                              desktops, NULL);
+        xfce_posix_signal_handler_set_handler(SIGTERM,
+                                              xfdesktop_handle_quit_signals,
+                                              desktops, NULL);
+        xfce_posix_signal_handler_set_handler(SIGUSR1,
+                                              xfdesktop_handle_SIGUSR1,
+                                              desktops, NULL);
+    } else {
+        g_warning("Unable to set up POSIX signal handlers: %s", error->message);
+        g_error_free(error);
     }
-    
-    signal(SIGHUP, sighandler_cb);
-    signal(SIGINT, sighandler_cb);
-    signal(SIGTERM, sighandler_cb);
-    signal(SIGUSR1, sighandler_cb);
-    
+
     gtk_main();
     
     menu_cleanup();
@@ -400,7 +412,11 @@ main(int argc, char **argv)
         gtk_widget_destroy(desktops[i]);
     g_free(desktops);
     
-    settings_cleanup();
+    xfconf_shutdown();
+
+#ifdef HAVE_LIBNOTIFY
+    xfdesktop_notify_uninit();
+#endif
     
     return 0;
 }
